unit uRTF;

{$INCLUDE 'RTFConv.inc'}

interface

uses
  SysUtils, Classes,
  {$IFDEF USE_GROUPS} uRtfGroups, {$ENDIF}
  uRtfTokens;

type
  ERtfError = class(Exception);

  TRtf = class
    private
      fTokens: TRtfTokenList;
      {$IFDEF USE_GROUPS}
      fGroups: TRtfGroupList;
    function IsComment(ATokenIndex): boolean;
      {$ENDIF}
    protected
      function ReadToken(var AData: PChar; var ASize: integer): TRtfToken; virtual;
      procedure Tokenize(AData: PChar; ASize: integer); virtual;
      procedure TokenizeGroup({$IFDEF USE_GROUPS} AGroup: TRtfGroup; AGroupList: TRtfGroupList; {$ENDIF} var AData: PChar; var ASize: integer); virtual;
      procedure SkipGroup(var ATokenIndex: integer {$IFDEF USE_GROUPS}; AGroup: TRtfGroup {$ENDIF}); virtual;
      function IsComment(ATokenIndex: integer): boolean; overload; virtual;
      function IsComment(ATokenIndex: integer; out ANextTokenType: TRtfTokenType; out ANextToken, ANextParam: string): boolean; overload; virtual;
      property Tokens: TRtfTokenList read fTokens;
      {$IFDEF USE_GROUPS}
      property Groups: TRtfGroupList read fGroups;
      {$ENDIF}
    public
      constructor Create; virtual;
      destructor Destroy; override;
      procedure Clear; virtual;
      procedure LoadFromStream(Stream: TStream); virtual;
      procedure LoadFromFile(const FileName: string); virtual;
      procedure SaveToStream(Stream: TStream); virtual;
      procedure SaveToFile(const FileName: string); virtual;
    end;

implementation

function GetToken(const DataStart, DataEnd: PChar): string;
var n: integer;
begin
  n := Integer(DataEnd) - Integer(DataStart);
  if n = 0 then
    Result := ''
  else
    begin
    SetLength(Result, n);
    Move(DataStart^, Result[1], n);
    end;
end;

function GetRtfToken(var Data: PChar; var DataLength: integer; out Token, Parameter: string; out TokenType: TRtfTokenType): boolean;
const Characters = {$IFDEF CASE_INSENSITIVE_TOKENS} ['a'..'z', 'A'..'Z'] {$ELSE} ['a'..'z'] {$ENDIF} ;
var DataStart: PChar;
begin
  Result := False;
  DataStart := Data;
  Token := '';
  Parameter := '';
  TokenType := rtfError;
  if DataLength > 0 then
    case Data^ of
      '{': // Group start
        begin
        Inc(Data);
        Dec(DataLength);
        Token := GetToken(DataStart, Data);
        TokenType := rtfGroupStart;
        Result := True;
        end;
      '}': // Group end
        begin
        Inc(Data);
        Dec(DataLength);
        Token := GetToken(DataStart, Data);
        TokenType := rtfGroupEnd;
        Result := True;
        end;
      '\': // Control word, Control symbol
        begin
        Inc(Data);
        Dec(DataLength);
        if DataLength > 0 then
          if Data^ in ['a'..'z'] then // Control word
            begin
            Result := True;
            while (DataLength > 0) and (Data^ in Characters) do
              begin
              Inc(Data);
              Dec(DataLength);
              end;
            Token := GetToken(DataStart, Data);
            TokenType := rtfControlWord;
            if DataLength > 0 then
              case Data^ of
                ' ': // Space is a part of control word
                  begin
                  Inc(Data);
                  Dec(DataLength);
                  end;
                '-', '0'..'9': // Numeric parameter becomes a part of the control word
                  begin
                  DataStart := Data;
                  repeat
                    Inc(Data);
                    Dec(DataLength);
                  until (DataLength <= 0) or (not (Data^ in ['0'..'9']));
                  Parameter := GetToken(DataStart, Data);
                  TokenType := rtfControlWordWithParam;
                  if (DataLength > 0) and (Data^ = ' ') then
                    begin
                    Inc(Data);
                    Dec(DataLength);
                    end;
                  end;
                else // Terminates the control word but is not a part of it
                  begin
                  end;
                end;
            end
          else // Control symbol
            begin
            Inc(Data);
            Dec(DataLength);
            Token := GetToken(DataStart, Data);
            TokenType := rtfControlSymbol;
            Result := True;
            end;
        end;
      else // Plain text
        begin
        while (DataLength > 0) and (not (Data^ in ['\','{','}'])) do
          begin
          Inc(Data);
          Dec(DataLength);
          end;
        Token := GetToken(DataStart, Data);
        TokenType := rtfPlainText;
        Result := True;
        end;
      end;
end;

{ TRtf }

constructor TRtf.Create;
begin
  inherited Create;
  fTokens := TRtfTokenList.Create;
  {$IFDEF USE_GROUPS}
  fGroups := TRtfGroupList.Create;
  {$ENDIF}
end;

destructor TRtf.Destroy;
begin
  FreeAndNil(fTokens);
  {$IFDEF USE_GROUPS}
  FreeAndNil(fGroups);
  {$ENDIF}
  inherited;
end;

procedure TRtf.Clear;
begin
  Tokens.Clear;
  {$IFDEF USE_GROUPS}
  Groups.Clear;
  {$ENDIF}
end;

procedure TRtf.LoadFromFile(const FileName: string);
var S: TFileStream;
begin
  S := TFileStream.Create(FileName, fmOpenRead or fmShareDenyWrite);
  try
    LoadFromStream(S);
  finally
    S.Free;
    end;
end;

procedure TRtf.LoadFromStream(Stream: TStream);
var Buf: Pointer;
    Size: integer;
begin
  Clear;
  Size := Stream.Size - Stream.Position;
  if Size > 0 then
    begin
    GetMem(Buf, Size);
    try
      Stream.ReadBuffer(Buf^, Size);
      Tokenize(Buf, Size);
    finally
      FreeMem(Buf);
      end;
    end;
end;

procedure TRtf.SaveToFile(const FileName: string);
var S: TFileStream;
begin
  S := TFileStream.Create(FileName, fmCreate or fmShareDenyWrite);
  try
    SaveToStream(S);
  finally
    S.Free;
    end;
end;

procedure TRtf.SaveToStream(Stream: TStream);
var i, n: integer;
    s: string;
begin
  n := Pred(Tokens.Count);
  for i := 0 to n do
    with Tokens[i] do
      if TokenType in [rtfControlWord, rtfControlWordWithParam] then
        begin
        s := Token + Parameter;
        if i < n then
          if Tokens[i+1].TokenType in [rtfPlainText, rtfBinaryData] then
            s := s + ' ';
        Stream.WriteBuffer(s[1], Length(s));
        end
      else
        Stream.WriteBuffer(Token[1], Length(Token));
end;

procedure TRtf.Tokenize(AData: PChar; ASize: integer);
begin
  if ReadToken(AData, ASize).TokenType = rtfGroupStart then
    begin
    TokenizeGroup({$IFDEF USE_GROUPS} nil, Groups, {$ENDIF} AData, ASize);
    if ASize <> 0 then
      Raise ERtfError.Create('Invalid RTF File: Data after the last group-end.');
    end
  else
    Raise ERtfError.Create('Invalid RTF File: Must start with group-start.');
end;

procedure TRtf.TokenizeGroup({$IFDEF USE_GROUPS} AGroup: TRtfGroup; AGroupList: TRtfGroupList; {$ENDIF} var AData: PChar; var ASize: integer);
var Tok: TRtfToken;
    {$IFDEF USE_GROUPS}
    Grp: TRtfGroup;
    {$ENDIF}
    n: integer;
begin
  repeat
    Tok := ReadToken(AData, ASize);
    case Tok.TokenType of
      rtfGroupStart:
        begin
          {$IFDEF USE_GROUPS}
          Grp := TRtfGroup.Create;
          Grp.FirstToken := Tokens.Count; // First token after GroupStart
          AGroupList.Add(Grp);
          {$ENDIF}
          TokenizeGroup({$IFDEF USE_GROUPS} Grp, Grp.Groups, {$ENDIF} AData, ASize);
        end;
      rtfGroupEnd:
        begin
          {$IFDEF USE_GROUPS}
          if AGroup <> nil then
            AGroup.LastToken := Pred(Tokens.Count); // The closing curly brace
          {$ENDIF}
        end;
      rtfControlWordWithParam:
        begin
          // We need to take care of '\bin' control word
          if Tok.Token = '\bin' then
            begin
            n := StrToInt(Tok.Parameter);
            if n < 0 then
              Raise ERtfError.Create('Binary data length must be positive.')
            else if n > ASize then
              Raise ERtfError.Create('Unexpected end of file while reading binary data.')
            else
              begin
              Tok := TRtfToken.Create;
              SetLength(Tok.Token, n);
              Move(AData^, Tok.Token[1], n);
              Inc(AData, n);
              Dec(ASize, n);
              Tok.Parameter := '';
              Tok.TokenType := rtfBinaryData;
              Tokens.Add(Tok);
              end;
            end;
        end;
      end;
  until Tok.TokenType = rtfGroupEnd;
end;

function TRtf.ReadToken(var AData: PChar; var ASize: integer): TRtfToken;
var Tok, Par: string;
    TokType: TRtfTokenType;
begin
  if GetRtfToken(AData, ASize, Tok, Par, TokType) then
    begin
    Result := TRtfToken.Create(TokType, Tok, Par);
    Tokens.Add(Result);
    end
  else
    Raise ERtfError.Create('Invalid token in RTF file.');
end;

procedure TRtf.SkipGroup(var ATokenIndex: integer {$IFDEF USE_GROUPS}; AGroup: TRtfGroup {$ENDIF});
{$IFNDEF USE_GROUPS}
var Depth: integer;
{$ENDIF}
begin
  {$IFDEF USE_GROUPS}
    ATokenIndex := Succ(AGroup.LastToken);
    if ATokenIndex >= Tokens.Count then
      Raise ERtfError.Create('Unfinished block encountered.');
  {$ELSE}
    Depth := 1;
    while (Depth > 0) and (ATokenIndex < Tokens.Count) do
      begin
      case Tokens[ATokenIndex].TokenType of
        rtfGroupStart:
          Inc(Depth);
        rtfGroupEnd:
          Dec(Depth);
        end;
      Inc(ATokenIndex);
      end;
    if Depth > 0 then
      Raise ERtfError.Create('Unfinished block encountered.');
  {$ENDIF}
end;

function TRtf.IsComment(ATokenIndex: integer): boolean;
begin
  with Tokens[ATokenIndex] do
    Result := (TokenType = rtfControlSymbol) and (Token = '\*');
end;

function TRtf.IsComment(ATokenIndex: integer; out ANextTokenType: TRtfTokenType; out ANextToken, ANextParam: string): boolean;
begin
  with Tokens[ATokenIndex] do
    Result := (TokenType = rtfControlSymbol) and (Token = '\*');
  if Result then
    if Succ(ATokenIndex) < Tokens.Count then
      with Tokens[Succ(ATokenIndex)] do
        begin
        ANextTokenType := TokenType;
        ANextToken := Token;
        ANextParam := Parameter;
        end
    else
      begin
      ANextTokenType := rtfError;
      ANextToken := '';
      ANextParam := '';
      end;
end;

end.

